import json

import pytest
import requests
from octokit import Octokit


class TestClientMethods(object):

    def test_client_methods_are_lower_case(self):
        for client in Octokit().__dict__:
            assert all(method.islower() for method in getattr(Octokit(), client).__dict__)

    def test_method_has_doc_string(self):
        assert Octokit().authorization.get.__doc__ == 'Get a single authorization.'

    def test_method_has_name_string(self):
        assert Octokit().authorization.get.__name__ == 'get'

    def test_method_calls_requests(self, mocker):
        mocker.patch('requests.get')
        Octokit().authorization.get(id=1)
        assert requests.get.called
        assert requests.get.call_count == 1

    def test_has_required_method_parameters(self):
        with pytest.raises(AssertionError):
            Octokit().authorization.get()
        with pytest.raises(AssertionError):
            Octokit().authorization.get(id=None)

    def test_only_allows_valid_method_parameters(self):
        with pytest.raises(AssertionError):
            Octokit().authorization.get_grants(notvalid=1)

    def test_validate_method_parameters(self, mocker):
        mocker.patch('requests.get')
        Octokit().authorization.get(id=100)
        requests.get.assert_called_once_with(
            'https://api.github.com/authorizations/100', params={}, headers=Octokit().headers
        )

    def test_request_has_body_parameters(self, mocker):
        mocker.patch('requests.post')
        data = {'scopes': ['public_repo'], 'note': 'admin script'}
        Octokit().authorization.create(**data)
        requests.post.assert_called_once_with(
            'https://api.github.com/authorizations', data=json.dumps(data), headers=Octokit().headers
        )

    def test_must_include_required_body_parameters(self):
        data = {'gist_id': 'abc123'}
        with pytest.raises(AssertionError):
            Octokit().authorization.create(**data)

    def test_use_default_parameter_values(self, mocker):
        mocker.patch('requests.patch')
        headers = {'accept': 'application/vnd.github.squirrel-girl-preview', 'Content-Type': 'application/json'}
        data = {'state': 'open'}
        Octokit().issues.edit(owner='testUser', repo='testRepo', number=1)
        requests.patch.assert_called_once_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )

    def test_use_passed_value_instead_of_default_parameter_values(self, mocker):
        mocker.patch('requests.patch')
        headers = {'accept': 'application/vnd.github.squirrel-girl-preview', 'Content-Type': 'application/json'}
        data = {'state': 'closed'}
        Octokit().issues.edit(owner='testUser', repo='testRepo', number=1, **data)
        requests.patch.assert_called_once_with(
            'https://api.github.com/repos/testUser/testRepo/issues/1', data=json.dumps(data), headers=headers
        )

    def test_validate_enum_values(self):
        with pytest.raises(AssertionError):
            Octokit().issues.edit(owner='testUser', repo='testRepo', number=1, state='closeddddd')

    def test_non_default_params_not_in_the_url_for_get_requests_go_in_the_query_string(self, mocker):
        mocker.patch('requests.get')
        params = {'page': 2, 'per_page': '30'}
        Octokit().authorization.get_grant(id=404, page=2)
        requests.get.assert_called_once_with(
            'https://api.github.com/applications/grants/404', params=params, headers=Octokit().headers
        )



class TestOctokit(object):

    def test_can_instantiate_class(self):
        import octokit
        assert isinstance(octokit.Octokit, object)
        octo = octokit.Octokit()
        assert isinstance(octo, object)

    def test_can_import_class(self):
        from octokit import Octokit
        assert isinstance(Octokit, object)
        octokit = Octokit()
        assert isinstance(octokit, object)

    def test_clients_are_lower_case(self):
        from octokit import Octokit
        assert all(client.islower() for client in Octokit.__dict__)



from octokit import webhook


class TestWebhook(object):

    def test_can_verify_webhook(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = {}
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events)

    def test_can_filter_webhook_events(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = {}
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events) is False

    def test_must_specify_events_to_allow(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = {}
        secret = 'secret'
        assert webhook.verify(headers, payload, secret) is False

    def test_can_specify_all_events(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = {}
        secret = 'secret'
        events = ['*']
        assert webhook.verify(headers, payload, secret, events=events)

    def test_only_known_events_are_valid(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'pushy',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958'
        }
        payload = {}
        secret = 'secret'
        events = ['pushy']
        assert webhook.verify(headers, payload, secret, events=events) is False

    def test_delivery_guids_must_be_valid_guids(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': 'not-a-guid'
        }
        payload = {}
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events) is False

    def test_can_verify_user_agent(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hookshot/',
        }
        payload = {}
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events, verify_user_agent=True)

    def test_verifies_user_agent(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hooks',
        }
        payload = {}
        secret = 'secret'
        events = ['push']
        assert webhook.verify(headers, payload, secret, events=events, verify_user_agent=True) is False

    def test_verify_ping_event(self):
        headers = {
            'X-Hub-Signature': 'sha1=02ad50e35622a4cf597de5fb86f4d59520f5adda',
            'X-GitHub-Event': 'ping',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hookshot/',
        }
        payload = {
            'hook': {
                'type': 'App',
                'id': 11,
                'active': True,
                'events': ['pull_request'],
                'app_id': 42,
            }
        }
        secret = 'secret'
        assert webhook.verify(headers, payload, secret, events=['*'], return_app_id=True) == 42

    def test_can_request_app_id_be_returned_on_non_ping_events(self):
        headers = {
            'X-Hub-Signature': 'sha1=5d61605c3feea9799210ddcb71307d4ba264225f',
            'X-GitHub-Event': 'push',
            'X-GitHub-Delivery': '72d3162f-cc78-11e3-81ab-4c9367dc0958',
            'User-Agent': 'GitHub-Hookshot/',
        }
        payload = {}
        secret = 'secret'
        assert webhook.verify(headers, payload, secret, events=['*'], return_app_id=True)



import json
import re
from collections import ChainMap

import requests
from octokit import utils


class Base(object):

    headers = {'accept': 'application/vnd.github.v3+json', 'Content-Type': 'application/json'}
    base_url = 'https://api.github.com'

    def _get_headers(self, definition):
        return dict(ChainMap(definition.get('headers', {}), self.headers))

    def _validate(self, kwargs, params):
        required_params = [k for k, v in params.items() if v.get('required')]
        for p in required_params:
            assert p in kwargs  # has all required
        for kwarg, value in kwargs.items():
            param_value = params.get(kwarg)
            assert param_value  # is a valid param but not necessarily required
            if param_value.get('enum'):
                assert value in param_value.get('enum')  # is a valid option of the enum
            if kwarg in required_params:
                assert value  # required param has a value

    def _form_url(self, values, _url):
        data_values = values.copy()
        for name, value in values.items():
            _url, subs = re.subn(f':{name}', str(value), _url)
            if subs != 0:
                data_values.pop(name)
        url = f'{self.base_url}{_url}'
        return url, data_values

    def _get_data(self, kwargs, params):
        for param, value in params.items():
            if value.get('default') and not kwargs.get(param):
                kwargs[param] = value.get('default')
        return kwargs


class Octokit(Base):

    def __init__(self):
        self._create(utils.get_json_data('rest.json'))

    def _create(self, definitions):
        for name, value in definitions.items():
            cls_name = utils.snake_case(name)
            client = self._create_client(cls_name, value)
            setattr(self, cls_name, client)

    def _create_client(self, name, methods):
        class_attributes = {}
        for _name, method in methods.items():
            method_name = utils.snake_case(str(_name))
            class_attributes.update({method_name: self._create_method(method_name, method)})
        bases = [object]
        cls = type(name, tuple(bases), class_attributes)
        return cls

    def _create_method(self, name, definition):

        def _api_call(*args, **kwargs):
            self._validate(kwargs, definition.get('params'))
            method = definition['method'].lower()
            requests_kwargs = {'headers': self._get_headers(definition)}
            url, data_kwargs = self._form_url(kwargs, definition['url'])
            data = self._get_data(data_kwargs, definition.get('params'))
            if method == 'get':
                requests_kwargs.update({'params': data})
            if method in ['post', 'patch', 'put', 'delete']:
                requests_kwargs.update({'data': json.dumps(data)})
            return getattr(requests, method)(url, **requests_kwargs)

        _api_call.__name__ = name
        _api_call.__doc__ = definition['description']
        return _api_call
